#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, Tuple
import mimiqcircuits as mc


class GateDecl:
    """
    Simple declaration of gates using the @gatedecl decorator.

    Examples:

        **First way**

        >>> from symengine import symbols
        >>> from mimiqcircuits import *

        Define symbols for the gate arguments

        >>> x, y = symbols('x y')

        Declare a gate using the @gatedecl decorator

        >>> @gatedecl("ansatz")
        ... def ansatz(x):
        ...     c = Circuit()
        ...     c.push(GateX(), 0)
        ...     c.push(GateRX(x), 1)
        ...     return c

        Create a GateDecl object using the decorator

        >>> ansatz(x)
        ansatz(x)

        >>> ansatz(y)
        ansatz(y)

        Decompose

        >>> ansatz(x).decompose()
        2-qubit circuit with 2 instructions:
        ├── X @ q[0]
        └── RX(x) @ q[1]
        <BLANKLINE>

         >>> ansatz(2).decompose()
         2-qubit circuit with 2 instructions:
         ├── X @ q[0]
         └── RX(2) @ q[1]
         <BLANKLINE>

        **Second Way**

        >>> from symengine import *
        >>> from mimiqcircuits import *

        Define symbols for the gate arguments

        >>> x, y = symbols('x y')

        From a circuit, create a GateDecl object directly

        >>> c = Circuit()
        >>> c.push(GateXXplusYY(x,y), 0,1)
        2-qubit circuit with 1 instructions:
        └── XXplusYY(x, y) @ q[0,1]
        <BLANKLINE>
        >>> c.push(GateRX(x), 1)
        2-qubit circuit with 2 instructions:
        ├── XXplusYY(x, y) @ q[0,1]
        └── RX(x) @ q[1]
        <BLANKLINE>
        >>> gate_decl = GateDecl("ansatz", ('x','y'), c)
        >>> gate_decl
        gate ansatz(x, y) =
        ├── XXplusYY(x, y) @ q[0,1]
        └── RX(x) @ q[1]
        <BLANKLINE>
        >>> GateCall(gate_decl, (2,4))
        ansatz(2, 4)

        Decompose the GateCall into a quantum circuit

        >>> GateCall(gate_decl, (2,4)).decompose()
        2-qubit circuit with 2 instructions:
        ├── XXplusYY(2, 4) @ q[0,1]
        └── RX(2) @ q[1]
        <BLANKLINE>

        Add to Circuit

        >>> g = GateCall(gate_decl, (2,4))
        >>> c = Circuit()
        >>> c.push(g,10,22)
        23-qubit circuit with 1 instructions:
        └── ansatz(2, 4) @ q[10,22]
        <BLANKLINE>
    """

    _name = "GateDecl"

    def __init__(
        self, name: str, arguments: Tuple[str, ...], circuit: mc.Circuit
    ):
        if not all(isinstance(arg, str) for arg in arguments):
            raise TypeError("All GateDecl arguments must be strings.")

        if len(circuit) == 0:
            raise ValueError("GateDecl cannot be defined from empty circuits.")

        for inst in circuit:
            # check if the instruction operation is from the Gate operation
            if not isinstance(inst.get_operation(), mc.Gate):
                raise ValueError("GateDecl instructions must be gates.")
            if inst.num_bits() != 0 or inst.num_zvars() != 0:
                raise ValueError("GateDecl instructions cannot have bits or zvars.")
            if inst.num_qubits() == 0:
                raise ValueError("GateDecl instructions must have qubits.")

        nq = circuit.num_qubits()
        np = len(arguments)
        super().__init__()

        self.name = name
        self.arguments = arguments
        self.circuit = circuit

    @property
    def num_qubits(self):
        return self.circuit.num_qubits()

    @num_qubits.setter
    def num_qubits(self, value):
        raise ValueError("Cannot set num_qubits. Read only parameter.")

    @property
    def num_bits(self):
        return self.circuit.num_bits()

    @num_bits.setter
    def num_bits(self, value):
        raise ValueError("Cannot set num_bits. Read only parameter.")

    @property
    def num_zvars(self):
        return self.circuit.num_zvars()

    @num_zvars.setter
    def num_zvars(self, value):
        raise ValueError("Cannot set num_zvars. Read only parameter.")

    def __str__(self):
        result = f"gate {self.name}({', '.join(map(str, self.arguments))}) =\n"
        N = len(self.circuit)

        for i, inst in enumerate(self.circuit):
            if i == N - 1:
                result += f"└── {inst}\n"
            else:
                result += f"├── {inst}\n"

        return result

    def __repr__(self):
        return str(self)


class GateCall(mc.Gate):
    """GateCall"""

    _name = "GateCall"
    _num_bits = 0
    _num_qubits = None

    def __init__(self, decl: GateDecl, args: Tuple[float, ...]):
        if len(args) != len(decl.arguments):
            raise ValueError("Wrong number of arguments for GateCall.")

        self._num_qubits = decl.num_qubits
        self._decl = decl
        self._args = args
        self._qregsizes = [self._num_qubits]

    def _matrix(self):
        pass

    def decompose(self):
        circ = mc.Circuit()
        d = dict(zip(self._decl.arguments, self._args))
        for inst in self._decl.circuit:
            op = inst.operation.evaluate(d)
            qubits = [i for i in inst.get_qubits()]
            bits = [i for i in inst.get_bits()]
            zvars = [i for i in inst.get_zvars()]
            circ.push(op, *qubits, *bits, *zvars)
        return circ

    def _decompose(self, circ, qubits, bits, zvars):
        d = dict(zip(self._decl.arguments, self._args))

        if len(qubits) != self._decl.num_qubits:
            raise ValueError("Wrong number of qubits for GateCall.")

        if len(bits) != self._decl.num_bits:
            raise ValueError("Wrong number of bits for GateCall.")

        if len(zvars) != self._decl.num_zvars:
            raise ValueError("Wrong number of zvars for GateCall.")

        for inst in self._decl.circuit:
            op = inst.operation.evaluate(d)

            # Map the instruction qubits to the actual qubits in the circuit
            inst_qubits = [qubits[q] for q in inst.get_qubits()]

            # Push the operation with the mapped qubits into the circuit
            circ.push(op, *inst_qubits, *bits, *zvars)

        return circ

    def __str__(self):
        return f"{self._decl.name}"

    def evaluate(self, d):
        new_args = [
            arg.subs(d) if hasattr(arg, "subs") else arg
            for arg in self._args
        ]
        return type(self)(self._decl, tuple(new_args))


def gatedecl(name):
    def decorator(func):
        def wrapper(*args, **kwargs):
            circ = func(*args, **kwargs)

            if not isinstance(circ, mc.Circuit):
                raise ValueError("GateDecl must return a Circuit object.")

            arguments = [str(arg) for arg in args] + [
                f"{key}={value}" for key, value in kwargs.items()
            ]

            decl = GateDecl(name, arguments, circ)
            return GateCall(decl, args)

        return wrapper

    return decorator


__all__ = ["GateCall", "GateDecl", "gatedecl"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from symengine import I, cos, sin, pi, eye, Matrix


def cis(x):
    return cos(x) + I * sin(x)


def cispi(x):
    return cos(pi * x) + I * sin(pi * x)


def cospi(x):
    return cos(pi * x)


def sinpi(x):
    return sin(pi * x)


def pmatrixpi(lmbda):
    return Matrix([[1, 0], [0, cispi(lmbda)]])


def pmatrix(lmbda):
    return Matrix([[1, 0], [0, cis(lmbda)]])


def ctrl(mat):
    """
    Returns the controlled version of a given 2x2 matrix.

    Args:
        mat (numpy.ndarray): A nxn matrix to be controlled.

    Returns:
        numpy.ndarray: A 2n x 2n controlled matrix.
    """
    dim = mat.shape[0]
    m = eye(2 * dim)

    # TODO: can it be done via slicing?
    for i in range(dim):
        for j in range(dim):
            m[i + dim, j + dim] = mat[i, j]
    return m


def gphase(lmbda):
    return cis(lmbda)


def gphasepi(lmbda):
    return cispi(lmbda)


def umatrix(theta, phi, lmbda, gamma=0.0):
    costheta2 = cos(theta / 2)
    sintheta2 = sin(theta / 2)
    return Matrix(
        [
            [cis(gamma) * costheta2, -cis(lmbda + gamma) * sintheta2],
            [cis(phi + gamma) * sintheta2, cis(phi + lmbda + gamma) * costheta2],
        ]
    )


def umatrixpi(theta, phi, lmbda, gamma=0):
    costheta2 = cospi(theta / 2)
    sintheta2 = sinpi(theta / 2)
    return Matrix(
        [
            [cispi(gamma) * costheta2, -cispi(lmbda + gamma) * sintheta2],
            [cispi(phi + gamma) * sintheta2, cispi(phi + lmbda + gamma) * costheta2],
        ]
    )


def rxmatrixpi(theta):
    return umatrixpi(theta, -1 / 2, 1 / 2)


def rxmatrix(theta):
    return rxmatrixpi(theta / pi)


def rymatrixpi(theta):
    return umatrixpi(theta, 0, 0)


def rymatrix(theta):
    return rymatrixpi(theta / pi)


def rzmatrixpi(lmbda):
    return gphasepi(-lmbda / 2) * umatrixpi(0, 0, lmbda)


def rzmatrix(lmbda):
    return rzmatrixpi(lmbda / pi)



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.proto import bitvector_pb2
from mimiqcircuits.proto.qcsrproto import bitvec_to_bytes, bytes_to_bitvec


def toproto_bitvector(bv):
    return bitvector_pb2.BitVector(len=len(bv), data=bytes(bitvec_to_bytes(bv)))


def fromproto_bitvector(bv):
    return bytes_to_bitvec(bv.data, bv.len)



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.operations.gates.standard.cpauli import GateCX
from symengine import Matrix
import mimiqcircuits as mc


class GateDCX(mcg.Gate):
    r"""Two qubit double-CNOT gate.

    A two qubit Clifford gate consisting of two back-to-back CNOTs with
    alternate controls.

    **Matrix representation:**

    .. math::
        \operatorname{DCX} =\begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            0 & 0 & 0 & 1 \\
            0 & 1 & 0 & 0
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateDCX()
        DCX
        >>> GateDCX().matrix()
        [1.0, 0, 0, 0]
        [0, 0, 1.0, 0]
        [0, 0, 0, 1.0]
        [0, 1.0, 0, 0]
        <BLANKLINE>
        >>> c = Circuit().push(GateDCX(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── DCX @ q[0,1]
        <BLANKLINE>
        >>> GateDCX().power(2), GateDCX().inverse()
        (DCX†, DCX†)
        >>> GateDCX().decompose()
        2-qubit circuit with 2 instructions:
        ├── CX @ q[0], q[1]
        └── CX @ q[1], q[0]
        <BLANKLINE>
    """

    _num_qubits = 2
    _qregsizes = [2]
    _name = "DCX"

    def _matrix(self):
        return Matrix(
            [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                [0.0, 1.0, 0.0, 0.0],
            ]
        )

    def _power(self, p):
        pmod = p % 3

        if pmod == 1:
            return self

        if pmod == 2:
            return mc.Inverse(self)

        if pmod == 0:
            return mc.GateID().parallel(2)

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits
        circ.push(GateCX(), a, b)
        circ.push(GateCX(), b, a)
        return circ



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from typing import List, Union
import symengine as se
import mimiqcircuits as mc
import sympy as sp
from mimiqcircuits.operations.krauschannel import krauschannel


def is_valid_power_of_2(n):
    """Check if a number is a power of 2."""
    return n > 0 and (n & (n - 1)) == 0


def _is_approx_identity(matrix, size, rtol=1e-12, atol=1e-12):
    """Check if the matrix is approximately the identity matrix using relative and absolute tolerance."""
    identity_matrix = se.eye(size)
    for i in range(size):
        for j in range(size):
            diff = abs(matrix[i, j] - identity_matrix[i, j])
            # Use both relative tolerance (rtol) and absolute tolerance (atol)
            if diff > atol + rtol * abs(identity_matrix[i, j]):
                return False
    return True


class Kraus(krauschannel):
    r"""Kraus(kmatrices).

    Custom N-qubit Kraus channel specified by a list of 2^N x 2^N Kraus matrices.

    A Kraus channel is defined by:

    .. math::
        \mathcal{E}(\rho) = \sum_k E_k \rho E_k^\dagger,

    where :math:`E_k` are Kraus matrices that need to fulfill :math:`\sum_k E_k^\dagger E_k = I`.

    If the Kraus matrices are all proportional to unitaries, use :func:`MixedUnitary` instead.

    The Kraus matrices are defined in the computational basis in the usual textbook order
    (the first qubit corresponds to the left-most qubit).
    For 1 qubit, we have :math:`|0\rangle`, :math:`|1\rangle`.
    For 2 qubits, we have :math:`|00\rangle`, :math:`|01\rangle`, :math:`|10\rangle`, :math:`|11\rangle`.

    **Note:** Currently, only 1 and 2-qubit custom Kraus channels are supported.

    See Also:
        :class:`MixedUnitary`
        :class:`GateCustom`

    Parameters:
        kmatrices (list): List of :math:`2^N \times 2^N` complex matrices. The number of qubits is equal to :math:`N`.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> c = Circuit()
        >>> c.push(Kraus([Matrix([[1, 0], [0, sqrt(0.9)]]), Matrix([[0, sqrt(0.1)], [0, 0]])]), 0)
        1-qubit circuit with 1 instructions:
        └── Kraus(Operator([[1, 0], [0, 0.948683298050514]]), Operator([[0, 0.316227766016838], [0, 0]])) @ q[0]
        <BLANKLINE>

        Kraus Matrices

        >>> g=Kraus([Matrix([[1, 0], [0, sqrt(0.9)]]), Matrix([[0, sqrt(0.1)], [0, 0]])])
        >>> g.krausmatrices()
        [[1.0, 0]
        [0, 0.948683298050514]
        , [0, 0.316227766016838]
        [0, 0]
        ]

        # Example 2: Kraus with Projector0 and Projector1

        >>> c = Circuit()
        >>> c.push(Kraus([Projector0(), Projector1()]), 0)
        1-qubit circuit with 1 instructions:
        └── Kraus(P₀(1), P₁(1)) @ q[0]
        <BLANKLINE>

        # Example 3: Kraus with a matrix and Projector1

        >>> c = Circuit()
        >>> c.push(Kraus([Matrix([[1, 0], [0, 0]]), Projector1()]), 0)
        1-qubit circuit with 1 instructions:
        └── Kraus(Operator([[1, 0], [0, 0]]), P₁(1)) @ q[0]
        <BLANKLINE>
    """

    _name = "Kraus"
    _parnames = ()

    def __init__(
        self,
        E: List[Union[se.Matrix, np.ndarray, sp.Matrix, mc.AbstractOperator]],
        rtol=1e-12,
    ):
        if len(E) == 0:
            raise ValueError("Vector of Kraus operators cannot be empty")

        Es = []
        for Ek in E:
            if isinstance(Ek, mc.AbstractOperator):
                Es.append(Ek)
            elif isinstance(Ek, (se.Matrix, np.ndarray, sp.Matrix)):
                Es.append(mc.Operator(self._convert_to_matrix(Ek)))
            else:
                raise TypeError(f"Unsupported operator type: {type(Ek)}")

        # Determine the dimension of the first operator
        dim = self._determine_dim(Es[0])

        # Ensure all matrices or operators have the same dimension
        for Ek in Es:
            matrix_dim = Ek.matrix().rows
            if matrix_dim != dim:
                raise ValueError("All matrices must have the same dimensions.")

        if not is_valid_power_of_2(dim):
            raise ValueError("Operator dimension must be 2^n.")

        self.N = Es[0].num_qubits
        if self.N < 1:
            raise ValueError("Cannot define a 0-qubit custom noise channel")
        if self.N > 2:
            raise ValueError(
                "Custom noise channels larger than 2 qubits are not supported"
            )

        # Perform the sum of E_k^† E_k = I check directly here
        M = 1 << self.N  # Equivalent to 2^N

        if not any(
            any(not element.is_number for element in op.matrix()) for op in Es
        ):
            # Perform normalization check only for purely numeric matrices
            ksum = sum(
                np.array(op.matrix().tolist(), dtype=np.complex128).conj().T
                @ np.array(op.matrix().tolist(), dtype=np.complex128)
                for op in Es
            )
            if not np.allclose(ksum, np.eye(M, dtype=np.complex128), rtol=1e-12):
                raise ValueError(
                    "List of Kraus matrices should fulfill \\sum_k E_k^\\dagger E_k = I."
                )

            # Check if ksum is approximately the identity matrix
            if not _is_approx_identity(ksum, M, rtol):
                raise ValueError(
                    "List of Kraus matrices should fulfill \\sum_k E_k^\\dagger E_k = I."
                )

        self.E = Es

        super().__init__()
        self._num_qubits = self.N
        self._parnames = ("E",)

    def evaluate(self, values: dict):
        """Evaluates symbolic parameters in Kraus matrices using a dictionary of values.

        Args:
            values (dict): A dictionary where keys are symbolic variables and values are the corresponding numerical values.

        Returns:
            Kraus: A new Kraus instance with evaluated matrices.
        """
        evaluated_E = []
        for op in self.E:
            if isinstance(op, mc.Operator):
                # Substitute symbolic parameters in the matrix
                matrix = op.matrix()
                evaluated_matrix = se.Matrix(
                    [[element.subs(values) for element in row] for row in matrix.tolist()]
                )
                evaluated_E.append(mc.Operator(evaluated_matrix))
            else:
                # If it's already an AbstractOperator, we just append it
                evaluated_E.append(op.evaluate(values))

        return Kraus(evaluated_E)


    def _convert_to_matrix(self, operator):
        """Converts matrices to symengine.Matrix, stripping imaginary parts if zero."""
        if isinstance(operator, (np.ndarray, sp.Matrix)):
            matrix = se.Matrix(operator.tolist())
        elif isinstance(operator, se.Matrix):
            matrix = operator
        else:
            raise TypeError(f"Unsupported matrix type: {type(operator)}")

        return matrix

    def _determine_dim(self, operator):
        """Determines the dimension of the operator or matrix."""
        return 1 << operator.num_qubits

    def krausoperators(self):
        """Returns the Kraus operators."""
        return self.E

    def __str__(self):
        """String representation, listing either AbstractOperators or matrix slices."""
        matrices_str = ", ".join(
            [
                f"[{matrix[0, 0]}, ... ,{matrix[-1, -1]}]"
                if isinstance(matrix, se.Matrix)
                else str(matrix)
                for matrix in self.E
            ]
        )
        return f"{self._name}({matrices_str})"

    def unwrappedkrausmatrices(self):
        """Returns the unwrapped Kraus matrices."""
        return self.krausmatrices()


__all__ = ["Kraus"]



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from typing import Union
import mimiqcircuits as mc
import symengine as se
from mimiqcircuits.operations.krauschannel import krauschannel
import sympy as sp


class PhaseAmplitudeDamping(krauschannel):
    r"""One-qubit phase amplitude damping noise channel.

    This channel is defined by:

    .. math::
        \mathcal{E}(\rho) =
        \begin{pmatrix}
            (1-\gamma)\rho_{00} + \gamma p & (1-2\beta)\sqrt{1-\gamma}\rho_{01} \\
            (1-2\beta)\sqrt{1-\gamma}\rho_{10} & (1-\gamma)\rho_{11} + (1-p)\gamma
        \end{pmatrix}

    Here, :math:`p, \gamma, \beta \in [0,1]`.

    This channel is equivalent to a `GeneralizedAmplitudeDamping(p, \gamma)` channel
    (see :class:`GeneralizedAmplitudeDamping`), followed by a `PauliZ(\beta)`
    channel (see :class:`PauliZ`).

    Use :func:`krausmatrices` to see a Kraus matrix representation of the channel.

    See Also:
        :class:`AmplitudeDamping`, :class:`GeneralizedAmplitudeDamping`,
        :class:`ThermalNoise`.

    Parameters:
        p (float): Probability parameter, must be in the range [0, 1].
        γ (float): Damping parameter, must be in the range [0, 1].
        β (float): Phase flip parameter, must be in the range [0, 1].

    Examples:
        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(PhaseAmplitudeDamping(0.1, 0.2, 0.3), 1)
        2-qubit circuit with 1 instructions:
        └── PhaseAmplitudeDamping(0.1, 0.2, 0.3) @ q[1]
        <BLANKLINE>
    """

    _name = "PhaseAmplitudeDamping"
    _num_qubits = 1
    _parnames = ()

    def __init__(
        self, p: Union[float, int], gamma: Union[float, int], beta: Union[float, int]
    ):
        if not isinstance(p, (se.Basic, sp.Basic)) and not (0 <= p <= 1):
            raise ValueError("Value of p must be between 0 and 1.")
        if not isinstance(gamma, (se.Basic, sp.Basic)) and not (0 <= gamma <= 1):
            raise ValueError("Value of gamma must be between 0 and 1.")
        if not isinstance(beta, (se.Basic, sp.Basic)) and not (0 <= beta <= 1):
            raise ValueError("Value of beta must be between 0 and 1.")

        self.p = p
        self.gamma = gamma
        self.beta = beta
        super().__init__()
        self._parnames = ("p", "gamma", "beta")

    def evaluate(self, d={}):
        attributes = ['p', 'gamma', 'beta']
        evaluated_values = {}

        for attr in attributes:
            value = getattr(self, attr)
            if hasattr(value, 'subs'):
                substituted_value = value.subs(d)
                evaluated_values[attr] = float(substituted_value.evalf()) if substituted_value.is_number else substituted_value
            else:
                evaluated_values[attr] = value

        # Extract evaluated values
        evaluated_p = evaluated_values['p']
        evaluated_gamma = evaluated_values['gamma']
        evaluated_beta = evaluated_values['beta']

        # Perform checks only if values are numeric
        if isinstance(evaluated_p, (float, int)) and not (0 <= evaluated_p <= 1):
            raise ValueError("Value of `p` must be between 0 and 1 after evaluation.")
        if isinstance(evaluated_gamma, (float, int)) and not (0 <= evaluated_gamma <= 1):
            raise ValueError("Value of `gamma` must be between 0 and 1 after evaluation.")
        if isinstance(evaluated_beta, (float, int)) and not (0 <= evaluated_beta <= 1):
            raise ValueError("Value of `beta` must be between 0 and 1 after evaluation.")

        # Return a new instance with evaluated parameters
        return PhaseAmplitudeDamping(evaluated_p, evaluated_gamma, evaluated_beta)

    def krausmatrices(self):
        return [se.Matrix(kraus.matrix().tolist()) for kraus in self.krausoperators()]

    def krausoperators(self):
        p = self.p
        gamma = self.gamma
        beta = self.beta
        K = np.sqrt(1 - gamma) * (1 - 2 * beta) / (1 - gamma * p)
        pref1 = np.sqrt(1 - gamma * p)
        pref2 = np.sqrt(1 - gamma * (1 - p) - (1 - gamma * p) * K**2)
        pref3 = np.sqrt(gamma * p)
        pref4 = np.sqrt(gamma * (1 - p))

        return [
            mc.DiagonalOp(pref1 * K, pref1),
            mc.Projector0(pref2),
            mc.SigmaMinus(pref3),
            mc.SigmaPlus(pref4),
        ]


    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"{self._name}{self.p,self.gamma,self.beta}"


class ThermalNoise(krauschannel):
    r"""One-qubit thermal noise channel.

    The thermal noise channel is equivalent to the :class:`PhaseAmplitudeDamping` channel,
    but it is parametrized instead as:

    .. math::
        \mathcal{E}(\rho) =
        \begin{pmatrix}
            e^{-\Gamma_1 t}\rho_{00} + (1-n_e)(1-e^{-\Gamma_1 t}) & e^{-\Gamma_2 t}\rho_{01} \\
            e^{-\Gamma_2 t}\rho_{10} & e^{-\Gamma_1 t}\rho_{11} + n_e(1-e^{-\Gamma_1 t})
        \end{pmatrix}

    where :math:`\Gamma_1 = 1/T_1` and :math:`\Gamma_2 = 1/T_2`, and the parameters must fulfill
    :math:`T_1 \geq 0`, :math:`T_2 \leq 2 T_1`, :math:`t \geq 0`, and :math:`0 \leq n_e \leq 1`.

    These parameters can be related to the ones used to define the :class:`PhaseAmplitudeDamping`
    channel through :math:`p = 1-n_e`, :math:`\gamma = 1-e^{-\Gamma_1 t}`, and
    :math:`\beta = \frac{1}{2}(1-e^{-(\Gamma_2-\Gamma_1/2)t})`.

    See Also:
        :class:`PhaseAmplitudeDamping`, :class:`AmplitudeDamping`,
        :class:`GeneralizedAmplitudeDamping`.

    Parameters:
        T₁ (float): Longitudinal relaxation rate, must be greater than or equal to 0.
        T₂ (float): Transversal relaxation rate, must be less than or equal to `2 * T₁`.
        t (float): Time duration of the gate, must be greater than or equal to 0.
        nₑ (float): Excitation fraction when in thermal equilibrium with the environment, must be in the range [0, 1].

    Examples:
        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(ThermalNoise(0.5, 0.6, 1.2, 0.3), 1)
        2-qubit circuit with 1 instructions:
        └── ThermalNoise(0.5, 0.6, 1.2, 0.3) @ q[1]
        <BLANKLINE>
    """

    _name = "ThermalNoise"
    _num_qubits = 1
    _parnames = ()

    def __init__(
        self,
        T1: Union[float, int],
        T2: Union[float, int],
        time: Union[float, int],
        ne: Union[float, int],
    ):
        if not isinstance(T1, (se.Basic, sp.Basic)) and (T1 < 0):
            raise ValueError("Value of T1 must be >= 0.")
        if not isinstance(T2, (se.Basic, sp.Basic)) and  (T2 > 2 * T1):
            raise ValueError("Value of T2 must be <= 2 * T1.")
        if not isinstance(time, (se.Basic, sp.Basic)) and  (time < 0):
            raise ValueError("Value of time must be >= 0.")
        if not isinstance(ne, (se.Basic, sp.Basic)) and not (0 <= ne <= 1):
            raise ValueError("Value of ne must be between 0 and 1.")

        self.T1 = T1
        self.T2 = T2
        self.time = time
        self.ne = ne
        super().__init__()
        self._parnames = ("T1", "T2", "time", "ne")

    def evaluate(self, d={}):
        # List of attributes to evaluate
        attributes = ['T1', 'T2', 'time', 'ne']
        evaluated_values = {}

        # Evaluate each attribute
        for attr in attributes:
            value = getattr(self, attr)
            if hasattr(value, 'subs'):
                # Substitute values using the dictionary and ensure substitution is applied
                substituted_value = value.subs(d)
                # Evaluate to a float if numeric, otherwise keep symbolic
                evaluated_values[attr] = float(substituted_value.evalf()) if substituted_value.is_number else substituted_value
            else:
                evaluated_values[attr] = value

        # Extract evaluated values
        evaluated_T1 = evaluated_values['T1']
        evaluated_T2 = evaluated_values['T2']
        evaluated_time = evaluated_values['time']
        evaluated_ne = evaluated_values['ne']

        # Perform checks only if values are numeric
        if isinstance(evaluated_T1, (float, int)) and evaluated_T1 < 0:
            raise ValueError("Value of `T1` must be >= 0 after evaluation.")
        if isinstance(evaluated_T2, (float, int)) and evaluated_T2 > 2 * evaluated_T1:
            raise ValueError("Value of `T2` must be <= 2 * T1 after evaluation.")
        if isinstance(evaluated_time, (float, int)) and evaluated_time < 0:
            raise ValueError("Value of `time` must be >= 0 after evaluation.")
        if isinstance(evaluated_ne, (float, int)) and not (0 <= evaluated_ne <= 1):
            raise ValueError("Value of `ne` must be between 0 and 1 after evaluation.")

        # Return a new instance with evaluated parameters
        return ThermalNoise(evaluated_T1, evaluated_T2, evaluated_time, evaluated_ne)


    def krausmatrices(self):
        return [se.Matrix(kraus.matrix().tolist()) for kraus in self.krausoperators()]

    def krausoperators(self):
        Gamma1 = 1 / self.T1
        Gamma2 = 1 / self.T2

        p = 1 - self.ne
        gamma = 1 - np.exp(-Gamma1 * self.time)
        beta = 0.5 * (1 - np.exp(-(Gamma2 - Gamma1 / 2) * self.time))

        pad = PhaseAmplitudeDamping(p, gamma, beta)
        return pad.krausoperators()

    @staticmethod
    def matrix(gate):
        return gate.matrix()

    @property
    def parnames(self):
        return self._parnames

    def __str__(self):
        return f"{self._name}{self.T1, self.T2, self.time, self.ne}"


__all__ = ["ThermalNoise", "PhaseAmplitudeDamping"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.operations.gates.standard.cpauli import GateCX
from mimiqcircuits.operations.gates.standard.hadamard import GateH
from mimiqcircuits.operations.gates.standard.s import GateS
from symengine import Matrix, I


class GateISWAP(mcg.Gate):
    r""" Two qubit ISWAP gate.

    See Also :func:`GateISWAPDG` and :func:`GateSWAP`

    **Matrix representation:**

    .. math::
        \operatorname{ISWAP} = \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 0 & i & 0 \\
            0 & i & 0 & 0 \\
            0 & 0 & 0 & 1
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateISWAP()
        ISWAP
        >>> GateISWAP().matrix()
        [1.0, 0, 0, 0]
        [0, 0, 0.0 + 1.0*I, 0]
        [0, 0.0 + 1.0*I, 0, 0]
        [0, 0, 0, 1.0]
        <BLANKLINE>
        >>> c = Circuit().push(GateISWAP(), 0, 1)
        >>> GateISWAP().power(2), GateISWAP().inverse()
        (ISWAP**2, ISWAP†)
        >>> GateISWAP().decompose()
        2-qubit circuit with 6 instructions:
        ├── S @ q[0]
        ├── S @ q[1]
        ├── H @ q[0]
        ├── CX @ q[0], q[1]
        ├── CX @ q[1], q[0]
        └── H @ q[1]
        <BLANKLINE>
    """

    _name = "ISWAP"

    _num_qubits = 2
    _qregsizes = [2]

    def _matrix(self):
        return Matrix([[1, 0, 0, 0], [0, 0, I, 0], [0, I, 0, 0], [0, 0, 0, 1]])

    def _decompose(self, circ, qubits, bits, zvars):
        c, t = qubits
        circ.push(GateS(), c)
        circ.push(GateS(), t)
        circ.push(GateH(), c)
        circ.push(GateCX(), c, t)
        circ.push(GateCX(), t, c)
        circ.push(GateH(), t)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.operations.utils import power_idempotent, control_one_defined
from mimiqcircuits.operations.gates.standard.phase import GateP
from mimiqcircuits.operations.gates.standard.u import GateU
import mimiqcircuits as mc
from symengine import pi, I, Matrix


class GateX(mcg.Gate):
    r""" Single qubit Pauli-X gate.

    **Matrix representation:**

    .. math::
        \operatorname{X} = \begin{pmatrix}
            0 & 1 \\
            1 & 0
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateX()
        X
        >>> GateX().matrix()
        [0, 1.0]
        [1.0, 0]
        <BLANKLINE>
        >>> c = Circuit().push(GateX(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── X @ q[0]
        <BLANKLINE>
        >>> GateX().power(2), GateX().inverse()
        (ID, X)
        >>> GateX().decompose()
        1-qubit circuit with 1 instructions:
        └── U(pi, 0, pi, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "X"

    _num_qubits = 1
    _qregsizes = [1]

    def inverse(self):
        return self

    def _power(self, p):
        pmod = p % 2

        # sqrt(X) = SX
        if pmod == 1 / 2:
            return mc.GateSX()

        # SX * SX^3 = X * X = ID => SX^3 = SXDG
        if pmod == 3 / 2:
            return mc.GateSXDG()

        # X^(2n) = ID
        # X^(2n + 1) = X
        return power_idempotent(self, p)

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCX(), mc.GateCCX(), mc.GateC3X())

    def _matrix(self):
        return Matrix([[0, 1], [1, 0]])

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(pi, 0, pi), q)
        return circ


class GateY(mcg.Gate):
    r"""Single qubit Pauli-Y gate.

    **Matrix representation:**

    .. math::
        \operatorname{Y} = \begin{pmatrix}
            0 & -i \\
            i & 0
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateY()
        Y
        >>> GateY().matrix()
        [0, -0.0 - 1.0*I]
        [0.0 + 1.0*I, 0]
        <BLANKLINE>
        >>> c = Circuit().push(GateY(), 0)
        >>> GateY().power(2), GateY().inverse()
        (ID, Y)
        >>> GateY().decompose()
        1-qubit circuit with 1 instructions:
        └── U(pi, (1/2)*pi, (1/2)*pi, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "Y"

    _num_qubits = 1
    _qregsizes = [1]

    def inverse(self):
        return self

    def _power(self, p):
        # Y^(2n) = ID
        # Y^(2n + 1) = Y
        return power_idempotent(self, p)

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCY())

    def _matrix(self):
        return Matrix([[0, -I], [I, 0]])

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(pi, pi / 2, pi / 2), q)
        return circ


class GateZ(mcg.Gate):
    r"""Single qubit Pauli-Z gate.

    **Matrix representation:**

    .. math::
        \operatorname{Z} = \begin{pmatrix}
            1 & 0 \\
            0 & -1
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateZ()
        Z
        >>> GateZ().matrix()
        [1.0, 0]
        [0, -1.0]
        <BLANKLINE>
        >>> c = Circuit().push(GateZ(), 0)
        >>> GateZ().power(2), GateZ().inverse()
        (ID, Z)
        >>> GateZ().decompose()
        1-qubit circuit with 1 instructions:
        └── P(pi) @ q[0]
        <BLANKLINE>
    """

    _name = "Z"

    _num_qubits = 1
    _qregsizes = [1]

    def inverse(self):
        return self

    def _power(self, p):
        pmod = p % 2

        # sqrt(Z) = S
        if pmod == 1 / 2:
            return mc.GateS()

        # S * S^3 = Z * Z = ID => S^3 = SDG
        if pmod == 3 / 2:
            return mc.GateSDG()

        # sqrt(S) = T
        if pmod == 1 / 4:
            return mc.GateT()

        # T * T^7 = S^2 * S^2 = Z * Z = ID => T^7 = TDG
        if pmod == 7 / 4:
            return mc.GateTDG()

        # Z^(2n) = ID
        # Z^(2n + 1) = Z
        return power_idempotent(self, p)

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCZ())

    def _matrix(self):
        return Matrix([[1, 0], [0, -1]])

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateP(pi), q)
        return circ
